/*
 *  Ray Trace Mandlebulb
 *
 *  Demonstrate using the CPU, OpenMP or CUDA to perform ray
 *  tracing of the Mandelbulb.  Note that many functions run
 *  on BOTH the CPU and GPU
 *
 *  Key bindings:
 *  m/M        Cycle computations between CPU, OpenMP and CUDA
 *  wasd       Move light
 *  +/-        Increase/decrease power (n)
 *  []         Increase/decrease maximum number of iterations
 *  space      Toggle shadows
 *  .          Toggle fast normals
 *  arrows     Change view angle
 *  PgDn/PgUp  Zoom in and out
 *  0          Reset view angle
 *  ESC        Exit
 */

using namespace std;

#include "CSCIx239.h"
#include "InitGPUcu.h"
#include <vector>
#include <cuda.h>
#include "Vec.h"

//  Global constants
#define Rmax    2
#define MaxDist 4
#define MaxStep 100

//  Struct with values needed for mandelbulb
class MandelBulb
{
   public:
      int   n;
      int   maxiter;
      bool  fast;
      bool  shadow;
      Vec3  L;
      Mat3  rot;
      float zoom;
      int   wid,hgt,dim;
};

//  Global variables
int mode=2;                 //  CPU or GPU
int th=0;                   //  Azimuth of view angle
int ph=0;                   //  Elevation of view angle
int Th=30;                  //  Azimuth of light angle
int Ph=30;                  //  Elevation of light angle
unsigned char* pixels=NULL; //  Pixel array
MandelBulb mb;              //  Mandelbulb structure
int nthread;                //  Threads per block
const char* text[] = {"CPU","OpenMP","CUDA"};

//
//  Mandlebulb distance estimate
//
__host__ __device__ inline float Step(const Vec3& C,const MandelBulb& mb)
{
   //  Skip one iteration (Z = 0 + C)
   Vec3  Z  = C;
   //  Compute |Z|
   float R  = sqrt(Z.x*Z.x+Z.y*Z.y+Z.z*Z.z);
   //  Initialize |Z'|
   float dR = 1;
   //  Loop maxiter times or until divergence
   int   i = mb.maxiter;
   while (R<Rmax && i--)
   {
      //  Convert to spherical
      float ph = asinf(Z.z/R);
      float th = atan2f(Z.y,Z.x);
      //  Compute derivative recursively n|Z^(n-1)|Z'
      float Rn = powf(R,mb.n-1);
      dR = Rn*dR*mb.n + 1;
      //  Compute R^n
      Rn *= R;
      //  Compute sin(n*ph) & cos(n*ph)
      float sinph,cosph;
      sincosf(mb.n*ph,&sinph,&cosph);
      //  Compute sin(n*th) & cos(n*th)
      float sinth,costh;
      sincosf(mb.n*th,&sinth,&costh);
      //  Compute Z^n+C
      Z = Rn*Vec3(cosph*costh,cosph*sinth,sinph) + C;
      //  Compute |Z|
      R = sqrt(Z.x*Z.x+Z.y*Z.y+Z.z*Z.z);
   }
   //  Evaluate G(c)/G'(c)
   return 0.5f*log(R)*R/dR;
}

//
//  Mandelbulb normal vector
//
__host__ __device__ inline Vec3 Normal(const Vec3 &C,const MandelBulb& mb)
{
   const float d = 1e-3;
   if (mb.fast)
   {
      float c  = Step(C,mb);
      float cx = Step(C+Vec3(d,0,0),mb);
      float cy = Step(C+Vec3(0,d,0),mb);
      float cz = Step(C+Vec3(0,0,d),mb);
      return normalize(Vec3(c-cx,c-cy,c-cz));
   }
   else
   {
      float cx = Step(C-Vec3(d,0,0),mb) - Step(C+Vec3(d,0,0),mb);
      float cy = Step(C-Vec3(0,d,0),mb) - Step(C+Vec3(0,d,0),mb);
      float cz = Step(C-Vec3(0,0,d),mb) - Step(C+Vec3(0,0,d),mb);
      return normalize(Vec3(cx,cy,cz));
   }
}

/*
 *  Adaptive ray marching
 */
__host__ __device__ inline bool RayMarchA(const Vec3& org, const Vec3& dir,const MandelBulb& mb,Vec3& point)
{
   const float eps = 1e-3;
   float dist  = 0;
   point = org;
   int k=MaxStep;
   //  Estimate distance until max distance or max steps
   while (dist<MaxDist && k--)
   {
      float step = Step(point,mb);
      point += step*dir;
      dist  += step;
      //  Stop of step is small
      if (fabs(step)<eps) return true;
   }
   return false;
}

//
//  Non-adaptive ray marching
//
__host__ __device__ inline bool RayMarchNA(Vec3 point,const Vec3& dir,const MandelBulb& mb)
{
   if (!mb.shadow) return false;
   float step = 0.1;
   float dist = 0;
   const float eps = 0.1*step;
   int k=MaxStep;
   //  Estimate distance until max distance or max steps
   while (dist<MaxDist && k--)
   {
      point += step*dir;
      dist  += step;
      step *= 1.1f; //increase step regularly (ie resolution decrease with distance)
      if (Step(point,mb)<eps) return true;
   }
   return false;
}

//
//  Ray trace with no secondary rays
//
__host__ __device__ Color RayTrace(Vec3 org,Vec3 dir,const MandelBulb& mb)
{
   //  Scalar intensity
   float f=0;
   Vec3  P;
   //  Test for intersection
   if (RayMarchA(org,dir,mb,P))
   {
      //  Add ambient
      f = 0.3;
      //  Compute normal vector
      Vec3 N = Normal(P,mb);
      //  Compute cosine with light vector
      float I = -(mb.L*N);
      //  If illuminated and not in shadow, add diffuse
      if (I>0 && !RayMarchNA(P,mb.L,mb)) f += 0.7*I;
   }
   //  Return the resulting color
   return Color(f,f,0);
}

//
//  Ray trace a pixel
//
__host__ __device__ void RayTracePixel(unsigned char* pix,int k,const MandelBulb& mb)
{
   //  Translate scalar index into pixel
   int i = k%mb.wid;
   int j = k/mb.wid;
   // Cast the intial ray from (x,y,1.1) down z axis
   Vec3 org(mb.zoom*(i-mb.wid/2) , mb.zoom*(j-mb.hgt/2) , 1.1);
   Vec3 dir(0,0,-1);
   Color col = RayTrace(mb.rot*org , mb.rot*dir , mb);
   //  Copy color to pixel array
   pix += 4*k;
   *pix++ = col.r>1 ? 255 : (unsigned int)(255*col.r);
   *pix++ = col.g>1 ? 255 : (unsigned int)(255*col.g);
   *pix++ = col.b>1 ? 255 : (unsigned int)(255*col.b);
   *pix++ = 255;
}

//
//  Set rotation matrix
//
void SetRot(void)
{
   float M[16];
   mat4identity(M);
   mat4rotate(M , ph,1,0,0);
   mat4rotate(M , th,0,1,0);
   //  Copy matrix to row vectors
   mb.rot.x.x = M[0]; mb.rot.x.y = M[4]; mb.rot.x.z = M[8];
   mb.rot.y.x = M[1]; mb.rot.y.y = M[5]; mb.rot.y.z = M[9];
   mb.rot.z.x = M[2]; mb.rot.z.y = M[6]; mb.rot.z.z = M[10];
}

//
//  Key pressed callback
//
void key(GLFWwindow* window,int key,int scancode,int action,int mods)
{
   //  Discard key releases (keeps PRESS and REPEAT)
   if (action==GLFW_RELEASE) return;

   //  Check for shift
   int shift = (mods & GLFW_MOD_SHIFT);

   //  Exit on ESC
   if (key == GLFW_KEY_ESCAPE)
      glfwSetWindowShouldClose(window,1);
   //  Reset view angle
   else if (key == GLFW_KEY_0)
      th = ph = 0;
   //  CPU/GPU
   else if (key == GLFW_KEY_M)
      mode = shift ? (mode+2)%3 : (mode+1)%3;
   //  Change power
   else if ((key==GLFW_KEY_KP_SUBTRACT || key==GLFW_KEY_MINUS) && mb.n>1)
      mb.n--;
   else if (key==GLFW_KEY_KP_ADD || key==GLFW_KEY_EQUAL)
      mb.n++;
   //  Change maximum iterations
   else if (key == GLFW_KEY_LEFT_BRACKET && mb.maxiter>0)
      mb.maxiter--;
   else if (key == GLFW_KEY_RIGHT_BRACKET)
      mb.maxiter++;
   //  Increase angle by 5 degrees
   else if (key == GLFW_KEY_A)
      Th += 5;
   //  Decrease angle by 5 degrees
   else if (key == GLFW_KEY_D)
      Th -= 5;
   //  Increase elevation by 5 degrees
   else if (key == GLFW_KEY_W)
      Ph += 5;
   //  Decrease elevation by 5 degrees
   else if (key == GLFW_KEY_S)
      Ph -= 5;
   //  Decrease elevation by 5 degrees
   else if (key == GLFW_KEY_SPACE)
      mb.shadow = !mb.shadow;
   else if (key == GLFW_KEY_PERIOD)
      mb.fast = !mb.fast;
   //  Right arrow key - increase angle by 5 degrees
   else if (key == GLFW_KEY_RIGHT)
      th += 5;
   //  Left arrow key - decrease angle by 5 degrees
   else if (key == GLFW_KEY_LEFT)
      th -= 5;
   //  Up arrow key - increase elevation by 5 degrees
   else if (key == GLFW_KEY_UP)
      ph += 5;
   //  Down arrow key - decrease elevation by 5 degrees
   else if (key == GLFW_KEY_DOWN)
      ph -= 5;
   //  Page Up key - increase zoom
   else if (key == GLFW_KEY_PAGE_UP)
      mb.zoom *= 0.9;
   //  Page Down key - decrease zoom
   else if (key == GLFW_KEY_PAGE_DOWN)
      mb.zoom *= 1.1;

   //  Keep angles to +/-360 degrees
   th %= 360;
   ph %= 360;
   SetRot();
}

//
//  Window resized callback
//
void reshape(GLFWwindow* window,int width,int height)
{
   //  Get framebuffer dimensions (makes Apple work right)
   glfwGetFramebufferSize(window,&width,&height);
   //  Window dimensions
   mb.wid = width;
   mb.hgt = height;
   mb.dim = 4*width*height;
   //  Allocate pixels size of window
   delete pixels;
   pixels = new unsigned char [mb.dim];
   //  Set the viewport to the entire window
   glViewport(0,0, width,height);
}

//
//  Initialize Mandelbulb
//
void Init(MandelBulb& mb)
{
   //  Mandelbulb parameters
   mb.n=6;             //  Power
   mb.maxiter=10;      //  Max iterations
   mb.shadow=true;     //  Shadows
   mb.fast=true;       //  Fast normals
   mb.zoom=0.005;      //  Zoom level
   //  Projection
   SetRot();
}

//
//  Ray trace the scene host/device interface
//
__global__ void RayTracePixDev(unsigned char* pix,const MandelBulb mb)
{
   unsigned int k = blockIdx.x*blockDim.x+threadIdx.x;
   if (k<mb.wid*mb.hgt)
      RayTracePixel(pix,k,mb);
}

//
//  Refresh display
//
void display(GLFWwindow* window)
{
   Elapsed();
   //  Set light direction
   mb.L.x = Sin(Th)*Cos(Ph);
   mb.L.y =         Sin(Ph);
   mb.L.z = Cos(Th)*Cos(Ph);

   //  CPU
   if (mode==0)
   {
      for (int k=0;k<mb.hgt*mb.wid;k++)
         RayTracePixel(pixels,k,mb);
   }
   //  OpenMP
   else if (mode==1)
   {
      #pragma omp parallel for
      for (int k=0;k<mb.hgt*mb.wid;k++)
         RayTracePixel(pixels,k,mb);
   }
   //  CUDA
   else
   {
      //  Allocate pixel array on device
      unsigned char* devpix;
      if (cudaMalloc((void**)&devpix,mb.dim)) Fatal("Cannot allocate device memory for pixels\n");

      //  Ray trace scene
      int nblock = mb.wid*mb.hgt/nthread;
      if (nblock*nthread<mb.wid*mb.hgt) nblock++;
      RayTracePixDev<<<nblock,nthread>>>(devpix,mb);

      //  Copy pixels from device to host
      if (cudaMemcpy(pixels,devpix,mb.dim,cudaMemcpyDeviceToHost)) Fatal("Cannot copy pixels from device to host\n");

      //  Free device memory
      cudaFree(devpix);
   }

   //  Time ray tracing
   double t = Elapsed();
   //  Blit scene to screen
   glWindowPos2i(0,0);
   glDrawPixels(mb.wid,mb.hgt,GL_RGBA,GL_UNSIGNED_BYTE,pixels);
   //  Display
   glWindowPos2i(5,5);
   Print("Size %dx%d Time %.3fs Angle %d,%d Power %d MaxIter %d Mode %s",mb.wid,mb.hgt,t,th,ph,mb.n,mb.maxiter,text[mode]);
   //  Flush
   glFlush();
   glfwSwapBuffers(window);
}

//
//  Main program with GLFW event loop
//
int main(int argc,char* argv[])
{
   //  Initialize GLFW
   GLFWwindow* window = InitWindow("Mandelblub",0,800,600,&reshape,&key);

   //  Initialize Mandelbulb
   Init(mb);
   //  Initialize CUDA
   nthread = InitGPU(1)/2;

   //  Event loop
   ErrCheck("init");
   while(!glfwWindowShouldClose(window))
   {
      //  Display
      display(window);
      //  Wait for something to happen
      glfwWaitEvents();
   }
   //  Shut down GLFW
   glfwDestroyWindow(window);
   glfwTerminate();
   return 0;
}
